#include "twn_textures_c.h"
#include "twn_util.h"
#include "twn_util_c.h"
#include "twn_engine_context_c.h"
#include "twn_workers_c.h"

#include <SDL2/SDL.h>
#include <physfs.h>
#include <physfsrwops.h>
#include <stb_ds.h>
#include <stb_rect_pack.h>
#include <stb_image.h>

#include <stdbool.h>

typedef struct {
    SDL_RWops *rwops;
    size_t size;
    size_t position;
} TextureLoadingContext;


static SDL_mutex *textures_load_mutex;
static struct TextureLoadRequest {
    Texture result; /* will be copied into cache when it's time, freeing us from locking */
    size_t index; /* index into cache->hash to fill */
    /* 0 = awaits processing */
    /* 1 = in processing */
    /* 2 = success */
    /* 3 = failure */
    uint8_t status;
} *texture_load_queue;


static int load_read_callback(void *user, char *data, int size) {
    TextureLoadingContext *context = user;
    int read = (int)SDL_RWread(context->rwops, data, 1, size);
    context->position += read;
    if (read == 0)
        CRY_SDL("Error in streamed texture load.");
    return read;
}

static void load_skip_callback(void *user, int n) {
    TextureLoadingContext *context = user;
    context->position += n;
    Sint64 result = SDL_RWseek(context->rwops, n, RW_SEEK_CUR);
    SDL_assert_always(result != -1);
}

static int load_eof_callback(void *user) {
    TextureLoadingContext *context = user;
    return context->position == context->size;
}


static SDL_Surface *missing_texture_surface;
static uint16_t missing_texture_id = TEXTURE_KEY_INVALID.id;

static SDL_Surface *gen_missing_texture_surface(void) {
    Uint32 rmask, gmask, bmask;

    #if SDL_BYTEORDER == SDL_BIG_ENDIAN
        rmask = 0xff000000;
        gmask = 0x00ff0000;
        bmask = 0x0000ff00;
    #else
        rmask = 0x000000ff;
        gmask = 0x0000ff00;
        bmask = 0x00ff0000;
    #endif

    SDL_LockMutex(textures_load_mutex);

    if (!missing_texture_surface) {
        int const dim = + TEXTURE_BORDER_REPEAT_SIZE * 2;
        uint8_t *data = SDL_malloc(dim * dim * 3);
        for (int y = 0; y < dim; ++y) {
            for (int x = 0; x < dim; ++x) {
                /* diagonal stripes, chosen so that asked pixel uvs are corresponding to final output */
                data[(y * dim + x) * 3 + 0] = (x / 2 + y / 2) % 2 == 0 ? 175 : 0;
                data[(y * dim + x) * 3 + 1] = (x / 2 + y / 2) % 2 == 0 ? 0 : 0;
                data[(y * dim + x) * 3 + 2] = (x / 2 + y / 2) % 2 == 0 ? 175 : 0;
            }
        }

        missing_texture_surface = SDL_CreateRGBSurfaceFrom(data, dim, dim,
                                                           3 * 8,
                                                           dim * 3,
                                                           rmask, gmask, bmask, 0);
    }

    SDL_UnlockMutex(textures_load_mutex);

    return missing_texture_surface;
}


SDL_Surface *textures_load_surface(const char *path, bool apply_border) {
    if (SDL_strncmp(path, "!", 1) == 0)
        goto GET_MISSING_TEXTURE;

    SDL_RWops *handle = PHYSFSRWOPS_openRead(path);
    if (handle == NULL)
        goto ERR_CANNOT_OPEN_FILE;

    TextureLoadingContext context = {
        .rwops = handle,
        .size = SDL_RWsize(handle),
    };

    stbi_io_callbacks callbacks = {
        .read = load_read_callback,
        .skip = load_skip_callback,
        .eof = load_eof_callback,
    };

    int width, height, channels;
    void *image_mem = stbi_load_from_callbacks(&callbacks, &context, &width, &height, &channels, 0);
    SDL_FreeRW(handle);
    if (!image_mem)
        goto ERR_CANNOT_READ_IMAGE;

    Uint32 rmask, gmask, bmask, amask;

    #if SDL_BYTEORDER == SDL_BIG_ENDIAN
        rmask = 0xff000000;
        gmask = 0x00ff0000;
        bmask = 0x0000ff00;
        amask = 0x000000ff;
    #else
        rmask = 0x000000ff;
        gmask = 0x0000ff00;
        bmask = 0x00ff0000;
        amask = 0xff000000;
    #endif

    SDL_Surface* surface = SDL_CreateRGBSurfaceFrom(image_mem, width, height,
                                                    channels * 8,
                                                    width * channels,
                                                    rmask, gmask, bmask, amask);
    if (surface == NULL)
        goto ERR_CANNOT_CREATE_SURFACE;

    /* TODO: investigate possibility of growing 1px border on stbi side, reducing the overhead (right now texture is held 2 times in memory) */
    /* use in atlases introduces seams on filtering, add 1px padding, growing the resulting */
    if (apply_border && (width < 2048 && height < 2048)) {
        SDL_Surface* border = SDL_CreateRGBSurface(0,
                                                   width + TEXTURE_BORDER_REPEAT_SIZE * 2,
                                                   height + TEXTURE_BORDER_REPEAT_SIZE * 2,
                                                   channels * 8,
                                                   rmask, gmask, bmask, amask);
        if (surface == NULL)
            goto ERR_CANNOT_CREATE_BORDER;

        /* main portion */
        SDL_SoftStretch(surface,
                        &(SDL_Rect){ .x = 0, .y = 0, .w = width, .h = height },
                        border,
                        &(SDL_Rect){ .x = TEXTURE_BORDER_REPEAT_SIZE, .y = TEXTURE_BORDER_REPEAT_SIZE, .w = width, .h = height });

        /* left border */
        SDL_SoftStretch(surface,
                        &(SDL_Rect){ .w = 1, .h = height },
                        border,
                        &(SDL_Rect){ .y = TEXTURE_BORDER_REPEAT_SIZE, .w = TEXTURE_BORDER_REPEAT_SIZE, .h = height });

        /* right border */
        SDL_SoftStretch(surface,
                        &(SDL_Rect){ .x = width - 1, .w = 1, .h = height },
                        border,
                        &(SDL_Rect){ .y = TEXTURE_BORDER_REPEAT_SIZE, .x = width + TEXTURE_BORDER_REPEAT_SIZE, .w = TEXTURE_BORDER_REPEAT_SIZE, .h = height });

        /* up border */
        SDL_SoftStretch(surface,
                        &(SDL_Rect){ .w = width, .h = 1 },
                        border,
                        &(SDL_Rect){ .x = TEXTURE_BORDER_REPEAT_SIZE, .w = width, .h = TEXTURE_BORDER_REPEAT_SIZE });

        /* bottom border */
        SDL_SoftStretch(surface,
                        &(SDL_Rect){ .y = height - 1, .w = width, .h = 1 },
                        border,
                        &(SDL_Rect){ .x = TEXTURE_BORDER_REPEAT_SIZE, .y = height + TEXTURE_BORDER_REPEAT_SIZE, .w = width, .h = TEXTURE_BORDER_REPEAT_SIZE });

        stbi_image_free(image_mem);
        SDL_FreeSurface(surface);
        surface = border;
    }

    SDL_SetSurfaceBlendMode(surface, SDL_BLENDMODE_NONE);
    SDL_SetSurfaceRLE(surface, true);

    return surface;

ERR_CANNOT_CREATE_BORDER:
    SDL_FreeSurface(surface);
ERR_CANNOT_CREATE_SURFACE:
    stbi_image_free(image_mem);

ERR_CANNOT_READ_IMAGE:
ERR_CANNOT_OPEN_FILE:
    /* something didn't worked out, use a stub texture */
    log_warn("Cannot open image: %s, using a stub", path);
GET_MISSING_TEXTURE:
    return gen_missing_texture_surface();
}


static SDL_Surface *create_surface(int width, int height) {
    Uint32 rmask, gmask, bmask, amask;

    #if SDL_BYTEORDER == SDL_BIG_ENDIAN
        rmask = 0xff000000;
        gmask = 0x00ff0000;
        bmask = 0x0000ff00;
        amask = 0x000000ff;
    #else
        rmask = 0x000000ff;
        gmask = 0x0000ff00;
        bmask = 0x00ff0000;
        amask = 0xff000000;
    #endif

    SDL_Surface *surface = SDL_CreateRGBSurface(0,
                                                width,
                                                height,
                                                TEXTURE_ATLAS_BIT_DEPTH,
                                                rmask,
                                                gmask,
                                                bmask,
                                                amask);

    SDL_SetSurfaceBlendMode(surface, SDL_BLENDMODE_NONE);
    SDL_SetSurfaceRLE(surface, true);

    return surface;
}


/* adds a new, blank atlas surface to the cache */
static void add_new_atlas(TextureCache *cache) {
    /* TODO: create a PBO surface if possible, reducing duplication */
    SDL_Surface *new_atlas = create_surface((int)ctx.texture_atlas_size, (int)ctx.texture_atlas_size);
    arrput(cache->atlas_surfaces, new_atlas);
    arrput(cache->atlas_textures, create_gpu_texture(TEXTURE_FILTER_NEAREAST, false, 4, (int)ctx.texture_atlas_size, (int)ctx.texture_atlas_size));
}


static void upload_texture_from_surface(GPUTexture texture, SDL_Surface *surface) {
    SDL_LockSurface(surface);

    upload_gpu_texture(texture, surface->pixels, surface->format->BytesPerPixel, surface->w, surface->h);

    SDL_UnlockSurface(surface);
}


static void recreate_current_atlas_texture(TextureCache *cache, int atlas_index) {
    profile_start("atlas recreation");

    /* TODO: should surfaces be freed after they cannot be referenced in atlas builing? */
    /*       for example, if full page of 64x64 tiles was already filled, there's no real reason to process them further */
    SDL_Surface *atlas_surface = cache->atlas_surfaces[atlas_index];

    /* blit the texture surfaces onto the atlas */
    for (size_t i = 0; i < shlenu(cache->hash); ++i) {
        /* skip all that aren't part of currently built one */
        if (cache->hash[i].value.atlas_index != atlas_index)
            continue;

        /* skip loners */
        if (cache->hash[i].value.loner_texture != 0)
            continue;

        SDL_BlitSurface(cache->hash[i].value.data,
                        NULL,
                        atlas_surface,
                        &(SDL_Rect){
                            .x = (int)cache->hash[i].value.srcrect.x,
                            .y = (int)cache->hash[i].value.srcrect.y,
                            .w = (int)cache->hash[i].value.srcrect.w,
                            .h = (int)cache->hash[i].value.srcrect.h,
                        });
    }

    /* texturize it! */
    upload_texture_from_surface(cache->atlas_textures[atlas_index], atlas_surface);

    profile_end("atlas recreation");
}


/* uses the textures currently in the cache to create an array of stbrp_rects */
static stbrp_rect *create_rects_from_cache(TextureCache *cache) {
    stbrp_rect *rects = NULL;
    bool missing_texture_used = false;
    for (size_t i = 0; i < shlenu(cache->hash); ++i) {
        if (cache->hash[i].value.loner_texture != 0)
            continue;

        /* only put it once */
        if (!missing_texture_used && cache->hash[i].value.data == missing_texture_surface)
            missing_texture_used = true;

        const SDL_Surface *surface_data = cache->hash[i].value.data;
        stbrp_rect new_rect = {
            .w = surface_data->w,
            .h = surface_data->h,
        };

        arrput(rects, new_rect);
    }
    return rects;
}


/* returns an array which contains a _copy_ of every unpacked rect in rects. */
/* each of these copies will have their original index in rects saved in */
/* their `id` field, which is an int. */
static stbrp_rect *filter_unpacked_rects(stbrp_rect *rects) {
    stbrp_rect *unpacked_rects = NULL;
    for (size_t i = 0; i < arrlenu(rects); ++i) {
        /* already packed */
        if (rects[i].was_packed)
            continue;

        arrput(unpacked_rects, rects[i]);
        /* stb_rect_pack mercifully gives you a free userdata int */
        /* the index is saved there so the original array can be updated later */
        unpacked_rects[arrlenu(unpacked_rects)-1].id = (int)i;
    }
    return unpacked_rects;
}


/* updates the original rects array with the data from packed_rects */
/* returns true if all rects were packed successfully */
static bool update_rects(TextureCache *cache, stbrp_rect *rects, stbrp_rect *packed_rects, int atlas_index) {
    /* !!! do not grow either of the arrays !!! */
    /* the reallocation will try to reassign the array pointer, to no effect. */
    /* see stb_ds.h */
    bool packed_all = true;

    for (size_t i = 0; i < arrlenu(packed_rects); ++i) {
        /* we can check if any rects failed to be packed right here */
        /* it's not ideal, but it avoids another iteration */
        if (!packed_rects[i].was_packed) {
            /* mark it as potentially filled in next atlas page */
            cache->hash[packed_rects[i].id].value.atlas_index = atlas_index + 1;
            packed_all = false;
            continue;
        }

        rects[packed_rects[i].id] = packed_rects[i];
        /* while the order of the elements in the hash map is unknown to us, */
        /* their equivalents in `rects` are in that same (unknown) order, which means */
        /* we can use the index we had saved to find the original texture struct */
        cache->hash[packed_rects[i].id].value.atlas_index = atlas_index;
    }

    return packed_all;
}


/* updates the atlas location of every rect in the cache */
static void update_texture_rects_in_atlas(TextureCache *cache, stbrp_rect *rects) {
    int r = 0;
    for (size_t i = 0; i < shlenu(cache->hash); ++i) {
        if (cache->hash[i].value.loner_texture != 0)
            continue;

        cache->hash[i].value.srcrect = (Rect) {
            .x = (float)rects[r].x,
            .y = (float)rects[r].y,
            .w = (float)rects[r].w,
            .h = (float)rects[r].h,
        };

        r++;
    }
}


void textures_cache_init(TextureCache *cache, SDL_Window *window) {
    SDL_zero(*cache);

    cache->window = window;
    sh_new_arena(cache->hash);

    cache->node_buffer = SDL_malloc(ctx.texture_atlas_size * sizeof *cache->node_buffer);
    textures_load_mutex = SDL_CreateMutex();

    add_new_atlas(cache);
}


void textures_cache_deinit(TextureCache *cache) {
    /* free atlas textures */
    for (size_t i = 0; i < arrlenu(cache->atlas_textures); ++i) {
        delete_gpu_texture(cache->atlas_textures[i]);
    }
    arrfree(cache->atlas_textures);

    /* free atlas surfaces */
    for (size_t i = 0; i < arrlenu(cache->atlas_surfaces); ++i) {
        SDL_FreeSurface(cache->atlas_surfaces[i]);
    }
    arrfree(cache->atlas_surfaces);

    /* free cache hashes */
    for (size_t i = 0; i < shlenu(cache->hash); ++i) {
        /* TODO: better to have field that stores the source of memory directly, ugh */
        if (missing_texture_surface != NULL && cache->hash[i].value.data->pixels == missing_texture_surface->pixels)
            SDL_free(cache->hash[i].value.data->pixels);
        else if (cache->hash[i].value.srcrect.w < 2048 && cache->hash[i].value.srcrect.h < 2048)
            (void)0; /* do nothing, memory owned by surface */
        else
            stbi_image_free(cache->hash[i].value.data->pixels);
        SDL_FreeSurface(cache->hash[i].value.data);
    }
    shfree(cache->hash);

    SDL_DestroyMutex(textures_load_mutex);
    arrfree(texture_load_queue);

    SDL_free(cache->node_buffer);
}


static enum TextureMode infer_texture_mode(SDL_Surface *surface) {
    const uint32_t amask = surface->format->Amask;
    if (amask == 0)
        return TEXTURE_MODE_OPAQUE;

    enum TextureMode result = TEXTURE_MODE_OPAQUE;

    SDL_LockSurface(surface);

    for (int i = 0; i < surface->w * surface->h; ++i) {
        /* TODO: don't assume 32 bit depth ? */
        Color color;
        SDL_GetRGBA(((uint32_t *)surface->pixels)[i], surface->format, &color.r, &color.g, &color.b, &color.a);

        if (color.a == 0)
            result = TEXTURE_MODE_SEETHROUGH;
        else if (color.a != 255) {
            result = TEXTURE_MODE_GHOSTLY;
            break;
        }
    }

    SDL_UnlockSurface(surface);

    return result;
}

/* offloads surface load and transparency detection from main thread */
bool textures_load_workers_thread(void) {
    /* try grabbing some work */
    ssize_t texture_id = -1;
    ssize_t queue_index = -1;
    char *path = NULL; /* copy of a key, as it's not stable in arena */

    SDL_LockMutex(textures_load_mutex);
    for (size_t i = 0; i < arrlenu(texture_load_queue); ++i) {
        if (texture_load_queue[i].status == 0) {
            texture_id = texture_load_queue[i].index;
            path = SDL_strdup(ctx.texture_cache.hash[texture_id].key);
            texture_load_queue[i].status = 1; /* mark as in process */
            queue_index = i;
            break;
        }
    }
    SDL_UnlockMutex(textures_load_mutex);
    /* nothing to do, bail */
    if (queue_index == -1)
        return false;

    SDL_assert(texture_id != -1 && queue_index != -1);

    SDL_Surface *const surface = textures_load_surface(path, true);
    SDL_free(path);

    Texture const response = {
        .data = surface,
        .mode = infer_texture_mode(surface),
    };

    SDL_LockMutex(textures_load_mutex);

    texture_load_queue[queue_index].result = response;
    texture_load_queue[queue_index].status = 2; /* mark success */

    /* reuse this id in the future, allowing for draw call merging */
    if (surface == missing_texture_surface && missing_texture_id == TEXTURE_KEY_INVALID.id)
        missing_texture_id = (uint16_t)texture_id;

    SDL_UnlockMutex(textures_load_mutex);

    return true;
}


static TextureKey textures_load(TextureCache *cache, const char *path) {
    /* at this point we assume that texture isn't loaded */

    /* place a dummy for future lookups to know it will be loaded */
    /* as well as a place for worker to fill in */
    shput(cache->hash, path, (Texture){0});

    /* append a new request, use stable indices */
    struct TextureLoadRequest const request = {
        .index = shlenu(cache->hash) - 1,
    };
    SDL_LockMutex(textures_load_mutex);
    arrpush(texture_load_queue, request);
    SDL_UnlockMutex(textures_load_mutex);

    /* signal work to do */
    workers_add_job();

    cache->is_dirty = true;

    /* report the newly created slot */
    return (TextureKey){ (uint16_t)shlenu(cache->hash) - 1 };
}


/* it's safe to access everything without lock after this returns true and no public api is possible to call */
static bool textures_load_workers_finished(void) {
    bool result = true;
    SDL_LockMutex(textures_load_mutex);
    for (size_t i = 0; i < arrlenu(texture_load_queue); ++i) {
        if (texture_load_queue[i].status == 0 || texture_load_queue[i].status == 1) {
            result = false;
            break;
        }
    }
    SDL_UnlockMutex(textures_load_mutex);
    return result;
}


void textures_update_atlas(TextureCache *cache) {
    if (!cache->is_dirty)
        return;

    while (!textures_load_workers_finished())
        SDL_Delay(1);

    /* collect results */
    for (size_t i = 0; i < arrlenu(texture_load_queue); ++i) {
        SDL_assert(texture_load_queue[i].status == 2);

        Texture response = texture_load_queue[i].result;

        /* it's a "loner texture," it doesn't fit in an atlas so it's not in one */
        if (response.data->w >= (int)ctx.texture_atlas_size || response.data->h >= (int)ctx.texture_atlas_size) {
            if (ctx.game.debug) {
                if (response.data->w > 2048 || response.data->h > 2048)
                    log_warn("Unportable texture dimensions for %s, use 2048x2048 at max", cache->hash[texture_load_queue[i].index].key);
                if (!is_power_of_two(response.data->w) || !is_power_of_two(response.data->h))
                    log_warn("Unportable texture dimensions for %s, should be powers of 2", cache->hash[texture_load_queue[i].index].key);
            }
            response.loner_texture = create_gpu_texture(TEXTURE_FILTER_NEAREAST, true, response.data->format->BytesPerPixel, response.data->w, response.data->h);
            upload_texture_from_surface(response.loner_texture, response.data);
            response.srcrect = (Rect) { .w = (float)response.data->w, .h = (float)response.data->h };

        }

        cache->hash[texture_load_queue[i].index].value = response;
    }

    arrsetlen(texture_load_queue, 0);

    stbrp_rect *rects = create_rects_from_cache(cache);
    int current_atlas_index = 0;

    /* we have to keep packing, and creating atlases if necessary, */
    /* until all rects have been packed. */
    /* ideally, this will not iterate more than once. */
    bool textures_remaining = true;
    while (textures_remaining) {
        /* this function makes a lot more sense if you read stb_rect_pack.h */
        stbrp_context pack_ctx; /* target info */
        stbrp_init_target(&pack_ctx,
                          (int)ctx.texture_atlas_size,
                          (int)ctx.texture_atlas_size,
                          cache->node_buffer,
                          (int)ctx.texture_atlas_size);

        stbrp_rect *rects_to_pack = filter_unpacked_rects(rects);
        stbrp_pack_rects(&pack_ctx, rects_to_pack, (int)arrlen(rects_to_pack));

        textures_remaining = !update_rects(cache, rects, rects_to_pack, current_atlas_index);
        arrfree(rects_to_pack); /* got what we needed */

        /* some textures couldn't be packed */
        if (textures_remaining) {
            update_texture_rects_in_atlas(cache, rects);
            recreate_current_atlas_texture(cache, current_atlas_index);

            /* need a new atlas for next time */
            add_new_atlas(cache);
            if (current_atlas_index >= cache->atlas_index) ++cache->atlas_index;
            ++current_atlas_index;
        }
    }

    update_texture_rects_in_atlas(cache, rects);
    recreate_current_atlas_texture(cache, current_atlas_index);

    cache->is_dirty = false;

    arrfree(rects);
}

TextureKey textures_get_key(TextureCache *cache, const char *path) {
    /* to prevent hashing errors */
    if (!path) path = "!";

    /* hash tables are assumed to be stable, so we just return indices */
    const ptrdiff_t texture = shgeti(cache->hash, path);

    /* load it if it isn't */
    if (texture == -1) {
        return textures_load(cache, path);
    } else
        return (TextureKey){ (uint16_t)texture };
}

int32_t textures_get_atlas_id(const TextureCache *cache, TextureKey key) {
    if (m_texture_key_is_valid(key)) {
        if (cache->hash[key.id].value.loner_texture != 0)
            return -cache->hash[key.id].value.loner_texture;
        else
            return cache->hash[key.id].value.atlas_index;
    } else {
        CRY("Texture lookup failed.",
            "Tried to get atlas id that isn't loaded.");
        return 0;
    }
}

Rect textures_get_srcrect(const TextureCache *cache, TextureKey key) {
    if (m_texture_key_is_valid(key)) {
        Rect const srcrect = cache->hash[key.id].value.srcrect;
        if (srcrect.w >= 2048 || srcrect.h >= 2048)
            return srcrect;
        else
            /* offset to not include border*/
            return (Rect){ .x = srcrect.x + TEXTURE_BORDER_REPEAT_SIZE,
                           .y = srcrect.y + TEXTURE_BORDER_REPEAT_SIZE,
                           .w = srcrect.w - TEXTURE_BORDER_REPEAT_SIZE * 2,
                           .h = srcrect.h - TEXTURE_BORDER_REPEAT_SIZE * 2 };
    } else {
        CRY("Texture lookup failed.",
            "Tried to get texture that isn't loaded.");
        return (Rect){ 0, 0, 0, 0 };
    }
}


Rect textures_get_dims(const TextureCache *cache, TextureKey key) {
    if (m_texture_key_is_valid(key)) {
        if (cache->hash[key.id].value.loner_texture != 0)
            return cache->hash[key.id].value.srcrect;
        else
            return (Rect){ .w = (float)ctx.texture_atlas_size, .h = (float)ctx.texture_atlas_size };
    } else {
        CRY("Texture lookup failed.",
            "Tried to get texture that isn't loaded.");
        return (Rect){ 0, 0, 0, 0 };
    }
}


void textures_bind(const TextureCache *cache, TextureKey key) {
    if (m_texture_key_is_valid(key)) {
        if (cache->hash[key.id].value.loner_texture == 0)
            bind_gpu_texture(cache->atlas_textures[cache->hash[key.id].value.atlas_index]);
        else
            bind_gpu_texture(cache->hash[key.id].value.loner_texture);
    } else if (key.id == 0) {
        CRY("Texture binding failed.",
            "Tried to get texture that isn't loaded.");
    }
}


/* TODO: alternative schemes, such as: array texture, fragment shader and geometry division */
/* TODO: a way to trigger upload before it's used */
void textures_bind_repeating(const TextureCache *cache, TextureKey key) {
    if (m_texture_key_is_valid(key)) {
        if (cache->hash[key.id].value.loner_texture == 0) {

            /* already allocated */
            if (cache->hash[key.id].value.repeating_texture != 0) {
                bind_gpu_texture(cache->hash[key.id].value.repeating_texture);
                return;
            }

            const Texture texture = cache->hash[key.id].value;

            const GPUTexture repeating_texture = create_gpu_texture(TEXTURE_FILTER_NEAREAST,
                                                                    false,
                                                                    texture.data->format->BytesPerPixel,
                                                                    texture.data->w,
                                                                    texture.data->h);

            SDL_LockSurface(texture.data);

            /* TODO: optional glCopyImageSubData support, abstracted as copy_gpu_texture() */
            upload_gpu_texture(repeating_texture,
                               texture.data->pixels,
                               texture.data->format->BytesPerPixel,
                               texture.data->w,
                               texture.data->h);

            SDL_UnlockSurface(texture.data);

            cache->hash[key.id].value.repeating_texture = repeating_texture;

            bind_gpu_texture(repeating_texture);

        } else
            bind_gpu_texture(cache->hash[key.id].value.loner_texture);

    } else if (key.id == 0) {
        CRY("Texture binding failed.",
            "Tried to get texture that isn't loaded.");
    }
}


TextureMode textures_get_mode(const TextureCache *cache, TextureKey key) {
    if (m_texture_key_is_valid(key)) {
        return cache->hash[key.id].value.mode;
    } else {
        CRY("Texture binding failed.",
            "Tried to get texture that isn't loaded.");
        return TEXTURE_MODE_GHOSTLY;
    }
}


size_t textures_get_num_atlases(const TextureCache *cache) {
    return cache->atlas_index + 1;    
}

void textures_reset_state(void) {
    textures_cache_deinit(&ctx.texture_cache);
    textures_cache_init(&ctx.texture_cache, ctx.window);
}
